<#
.SYNOPSIS
Adds, removes and checks events for a JobScheduler Master instance.

.DESCRIPTION
Events can be managed for a JobScheduler Master or Supervisor instance that implements event handling.

This script is provided for compatibility reasons when using the script .\bin\jobscheduler_event.cmd 
from a Windows console as it implements the same parameter interface. When using PowerShell directly
then it is recommended to use the respective cmdlets instead of this script:

* Action 'add': Add-JobSchedulerEvent
* Action 'remove': Remove-JobSchedulerEvent
* Action 'check': Get-JobSchedulerEvent

Should the JobScheduler Master or Supervisor not be accessible then events are stored in 
the local file %TEMP%\jobscheduler_events.log for later dequeueing. Subsequent calls to this
script will dequeue any previously stored events.

Starting from JobScheduler Release 1.11 connections should not be made directly to the JobScheduler Master 
or Supervisor instance but to the JOC Cockpit REST Web Service as this guarantees authentication and
authorization.

.PARAMETER EventClass
Specifies a common name for a set of events that enable event handlers to process multiple events of the
same class. For example, "daily_closing" could be an event class for jobs that should start once day-time
business processes have drawn to a close.

Specifies a unique identifier when used together with the -EventId parameter.

.PARAMETER EventId
An identifier for an event. Allows event handlers to react to events having a particular ID.

Specifies a unique identifier when used together with the -EventClass parameter. An event id is required to be
unique for the same event class.

.PARAMETER ExitCode
Specifies the exit code that is added to the event. Usually this signals the execution status of a job, 
however, you can assign a specific numeric exit code.

Without this parameter being used the $LastExitCode is implicitely assumed, i.e. the last exit code
that was provided from a command or program.

Consider that exit codes other than 0 signal failed execution. You can specify allowed exit codes
with the -AllowedExitCodes parameter.

Default: last script exit code or failed execution of previous command.

.PARAMETER AllowedExitCodes
Specifies a list of exit codes that signal that a job is considered as having run successfully. 
This is useful if job scripts provide return values in the form of exit codes and these codes
should not be considered as errors. When adding an event then any exit codes that match one of the
allowed exit codes are set to 0.

A range of allowed exit codes is specified by e.g -AllowedExitCodes 1..4 or -AllowedExitCodes 1,2,3,4.

.PARAMETER Job
Specifies the name of the job that is assigned to the event.

Default: the current job name if the cmdlet is used by a job.

.PARAMETER JobChain
Specifies the name of the job chain that is assigned to the event.

Default: the current job chain name if the cmdlet is used by a job chain.

.PARAMETER Order
Specifies the identifier of the order assigned to the event.

Default: the current order identification (order id) if the cmdlet is used by a job chain.

.PARAMETER MasterHost
Specifies the Job Scheduler Master instance hostname or ip address.

This parameter is exclusively used when connecting directly to a JobScheduler Master.
When using the -Url parameter for access to JOC Copckpit REST Web Services then this
parameter is not specified.

Default: If used with a job then the CLI will assign by default the JobScheduler Master that the job is running for.

.PARAMETER MasterPort
Specifies the Job Scheduler Master instance TCP port.

This parameter is exclusively used when connecting directly to a JobScheduler Master.
When using the -Url parameter for access to JOC Copckpit REST Web Services then this
parameter is not specified.

Default: If used with a job then the CLI will assign by default the port of the JobScheduler Master that the job is running for.

.PARAMETER SupervisorHost
Specifies a Job Scheduler Supervisor instance hostname or ip address.

Job Scheduler Master instances register with a JobScheduler Supervisor (if configured to do so)
in order to synchronize job configurations. The Supervisor instance receives events, executes the
event handler and starts jobs and job chains for registered JobScheduler Master instances.

This parameter is exclusively used when connecting directly to a JobScheduler Supervisor.
When using the -Url parameter for access to JOC Copckpit REST Web Services then this
parameter is not specified.

Default: If used with a job then the CLI will assign by default the JobScheduler Supervisor that the 
current JobScheduler Master is registered for and otherwise assign the JobScheduler Master.

.PARAMETER SupervisorPort
Specifies the Job Scheduler Master instance TCP port.

This parameter is exclusively used when connecting directly to a JobScheduler Supervisor
When using the -Url parameter for access to JOC Copckpit REST Web Services then this
parameter is not specified.

Default: If used with a job then the CLI will assign by default the port of the JobScheduler Master that the job is running for.

.PARAMETER SupervisorJobChain
Specifies the path of the job chain in the JobScheduler Master or Supervisor instance that implements the event
processor. 

Default: /sos/events/scheduler_event_service

.PARAMETER What
This parameter specifies the effective action that is mapped to use of one of the cmdlets:

* Action 'add': Add-JobSchedulerEvent
* Action 'remove': Remove-JobSchedulerEvent
* Action 'check': Get-JobSchedulerEvent

Default: add

.PARAMETER ExpirationDate
Specifies the point in time for which an event automatically expires: this parameter is considered by the
Event Processor. A specific point in time can be specified by use of the ISO date format, i.e. yyyy-MM-dd HH:mm:ss.
The value 'never' ensures that the event does not ever expire. 

The parameters -ExpirationDate, -ExpirationCycle, -ExpirationPeriod may not be used at the same time.

Default: 00:00 on the following day.

.PARAMETER ExpirationCycle
Specifies the time for which the event will expire in the current period.
Periods start at 00:00 and end at 24:00. An expiration cycle of 21:00 
specifies 9pm of the current cycle for event expiration.

Values are specified by use of the format HH:mm:ss.

The parameters -ExpirationDate, -ExpirationCycle, -ExpirationPeriod, -NoExpiration may not be used at the same time.

.PARAMETER ExpirationPeriod
Specifies the duration after which the event will expire in the current period.
Periods start at 00:00 and end at 24:00. An expiration period of 04:00 
specifies that the event will expire 4 hours starting from the current point in time.

Values are specified by use of the format HH:mm:ss.

The parameters -ExpirationDate, -ExpirationCycle, -ExpirationPeriod, -NoExpiration may not be used at the same time.

.PARAMETER XPath
This parameter is used exlusively if the -What parameter with a value "check" is used.

All events corresponding to the XPath expression specified when this parameter is set. Complex expressions
are possible and any attributes of an event can be addressed. This parameter allows complex queries to
be made, that would not be possible with the -EventClass, -EventId and -ExitCode parameters.

.PARAMETER Parameters
Allows additional parameters for event handlers to be specified. Parameters are created using name-value
pairs from a hashmap as e.g. @{'name1'='value1'; 'name2'='value2'}.

Parameter names can be freely chosen and event handlers configured to
take account of values handed over. Any number of parameters can be added.

.PARAMETER Url
Specifies the Url of the JOC Cockpit REST Web Service. This parameter is used alternatively to the -MasterHost
and -MasterPort parameters. 

For JobScheduler releases starting from 1.11 it is recommended to use -Url in order to force
authentication and authorizatio by the JOC Cockpit REST Web Service.

The URL consists of the protocol, host name and port, e.g. http://localhost:4446.

If this parameter is not specified then the value is checked from the $SCHEDULER_DATA/config/private/private.conf file.
If this file contains a setting such as 'joc.url = "http://localhost:4446"' then the JOC Cockpit REST Web Service is looked up from the localhost at port 4446.

.PARAMETER Account
Specifies the credentials for use of an account with the JOC Cockpit REST Web Service.

Any access to the web service requires authentication, therefore this parameter is required if the -Url parameter is used.

If this parameter is not specified then the value is checked from the $SCHEDULER_DATA/config/private/private.conf file.
If this file contains a setting such as 'joc.account = "root:root"' then for access to the JOC Cockpit REST Web Service 
the account 'root' and password 'root' will be used.

.PARAMETER JobSchedulerId
The JobSchedulerId identifies the JobScheduler Master instance that operates as the event handler.

When used with a job on a JobScheduler Master or Agent then by default the identification of the respective JobScheduler Master is used.

Default: Current JobScheduler Master

.PARAMETER Timeout
Specifies the timeout for a connection to the JOC Cockpit Web Service or to a JobScheduler Master.

If this parameter is not specified then the value is checked from the $SCHEDULER_DATA/config/private/private.conf file.
If this file contains a setting such as 'http.timeout = 5000' then the timeout will be set to the respective number of milliseconds.

Default: 15s

.OUTPUTS
This script returns the number of events that have been added, retrieved or removed.

.EXAMPLE
./jobscheduler_event.ps1 -e daily_closing -i 12345678 -u http://localhost:4446 -n root:root -b localhost_4444 

Creates an event by use of a connection to the JOC Cockpit REST Web Service.

The web service URL is specified with the -u option. The account and password are specified with the -n option.
The -b option takes the JobSchedulerID, i.e. the string that identifies the JobScheduler Master or Supervisor
that is running the event handler job chain.

.EXAMPLE
./jobscheduler_event.ps1 -e daily_closing -i 12345678 -h localhost -p 4444

Creates an event with the JobScheduler Master instance that runs on the localhost for port 4444. 

This type of connection has to be used up to release 1.10. For later releases use the -Url parameter
to connect to the JOC Cockpit REST Web Service.

.EXAMPLE
./jobscheduler_event.ps1 -e daily_closing -i 12345678

Creates an event with the specified event class and event id. As no -Url or -MasterHost/-MasterPort 
parameters have been used therefore the following locations are checked in the given precedence:

* the -Url value is looked up from a ./config/private/private.conf file.
* the -MasterHost/-MasterPort values are looked up from environment variables that are proviced by a Master/Agent.

Maps to use of the cmdlet:

Add-JobSchedulerEvent -EventClass daily_closing -EventId 12345678

.EXAMPLE
./jobscheduler_event.ps1 -e daily_clsoing -i 12345678 -l 1..4

Creates an event with the specified event class and event id. The exit code is implicitely added
from the global $LastExitCode value. Should the exit code be contained in the list of 
allowed exit codes then its value is set to 0.

Maps to use of the cmdlet:

Add-JobSchedulerEvent -EventClass daily_closing -EventId 12345678 -AllowedExitCodes 1..4

.EXAMPLE
./jobscheduler_event.ps1 -e daily_clsoing -i 12345678 -tp 04:00

Creates an event that will expire 4 hrs. from now.

Maps to use of the cmdlet:

Add-JobSchedulerEvent -EventClass daily_closing -EventId 12345678 -ExpirationPeriod 04:00

.EXAMPLE
./jobscheduler_event.ps1 -e daily_clsoing -i 12345678 -tc 21:00

Creates an event that will expire at 9pm of the current day.

Maps to use of the cmdlet:

Add-JobSchedulerEvent -EventClass daily_closing -EventId 12345678 -ExpirationCycle 21:00

.LINK
about_jobscheduler

#>

param
(
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('e')]
    [string] $EventClass,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('i')]
    [string] $EventId,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('x')]
    [int] $ExitCode,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('l')]
    [int[]] $AllowedExitCodes,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('j')]
    [string] $Job,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('c')]
    [string] $JobChain,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('o')]
    [string] $Order,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('h')]
    [string] $MasterHost,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('p')]
    [int] $MasterPort,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('s')]
    [string] $SupervisorHost,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('r')]
    [int] $SupervisorPort,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('k')]
    [string] $SupervisorJobChain,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('w')]
    [string] $What = 'add',
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('t')]
    [string] $ExpirationDate,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('tc')]
    [string] $ExpirationCycle,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('tp')]
    [string] $ExpirationPeriod,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('a')]
    [string] $XPath,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('d')]
    [string[]] $Parameters,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('u')]
    [Uri] $Url,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('n')]
    [string] $Account,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('b')]
    [string] $JobSchedulerId,
    [Parameter(Mandatory=$False,ValueFromPipeline=$False,ValueFromPipelinebyPropertyName=$True)][Alias('m')]
    [int] $Timeout
)

    $privateConfigFile = "$($env:SCHEDULER_DATA)/config/private/private.conf"
    $isPrivateConfigFile = Test-Path -type Leaf -Path $privateConfigFile

    # defaults
    if ( $Timeout )
    {
        Set-JobSchedulerOption -WebRequestTimeout $Timeout
    }

    if ( !$Url -And $isPrivateConfigFile )
    {
        $result = select-string -Path $privateConfigFile -Pattern "^joc\.url\s*=\s*""?([^""]*)""?"
        if ( $result.Matches -And $result.Matches.Groups.length -gt 0 )
        {
            $Url = $result.Matches.Groups[1].Value
        }
    }
    
    if ( $Url -And !$Account -And $isPrivateConfigFile )
    {            
        $result = select-string -Path $privateConfigFile -Pattern "^joc\.account\s*=\s*""?([^""]*)""?"
        if ( $result.Matches -And $result.Matches.Groups.length -gt 0 )
        {
            $Account = $result.Matches.Groups[1].Value
        }
    }

    if ( !$JobSchedulerId )
    {
        if ( $env:SCHEDULER_ID )
        {
            $JobSchedulerId = $env:SCHEDULER_ID
        }
    }

    if ( !$Url )
    {
        if ( !$SupervisorHost -And $env:SCHEDULER_SUPERVISOR_HOST )
        {
            $SupervisorHost = $env:SCHEDULER_SUPERVISOR_HOST
        }
        
        if ( !$SupervisorPort -And $env:SCHEDULER_SUPERVISOR_PORT )
        {
            $SupervisorPort = $env:SCHEDULER_SUPERVISOR_PORT
        }

        if ( !$MasterHost -And $env:SCHEDULER_HOST )
        {
            $MasterHost = $env:SCHEDULER_HOST
        }
        
        if ( !$MasterPort -And $env:SCHEDULER_TCP_PORT )
        {
            $MasterPort = $env:SCHEDULER_TCP_PORT
        }

        if ( !$SupervisorHost -And $MasterHost )
        {
            $SupervisorHost = $MasterHost
        }
        
        if ( !$SupervisorPort -And $MasterPort )
        {
            $SupervisorPort = $MasterPort
        }
    }

    if ( !$Timeout -And $isPrivateConfigFile )
    {
        $result = select-string -Path $privateConfigFile -Pattern "^http\.timeout\s*=\s*""?([^""]*)""?"
        if ( $result.Matches -And $result.Matches.Groups.length -gt 0 )
        {
            $Timeout = $result.Matches.Groups[1].Value
        }
    }
    
    # parameter checking
    if ( ($MasterHost -Or $MasterPort) -And $Url )
    {
        throw 'Only one of the parameters -Master(H)ost/-Master(P)ort and -(U)rl can be used'
    }

    if ( ($SupervisorHost -Or $SupervisorPort) -And $Url )
    {
        throw 'Only one of the parameters -(S)upervisorHost/-Supe(r)visorPort and -(U)rl can be used'
    }

    if ( (-Not $SupervisorHost -Or -Not $SupervisorPort) -And (-Not $MasterHost -Or -Not $MasterPort) -And -Not $Url )
    {
        throw 'Either the -(U)rl parameter or both parameters -Master(H)ost and -Master(P)ort or both parameters -(S)upervisorHost and -Supe(r)visorPort have to be specified'
    }
    
    if ( $What -ne 'add' -And $What -ne 'remove' -And $What -ne 'check' )
    {
        throw 'One of the values "add", "remove" or "check" has to be specified for the -(W)hat parameter'
    }

	if ( $What -eq 'add' -And !$EventClass )
	{
		throw 'Parameter -EventClass is required for adding events'
	}
	
	if ( $What -eq 'add' -And !$EventId )
	{
		throw 'Parameter -EventId is required for adding events'
	}
	
    # Option 1: Connection to JobScheduler Master (up to release 1.10)
    if ( $SupervisorHost -And $SupervisorPort )
    {
        $ms = Use-JobSchedulerMaster -Url http://$($SupervisorHost):$($SupervisorPort)
    }
    
    # Option 2: Connection to JOC Cockpit Web Service (starting from release 1.11)
    if ( $Url )
    {
        if ( !$JobSchedulerId )
        {
            throw 'Parameter -Jo(b)SchedulerId has to be specified if -(U)rl parameter is used'
        }
        
        if ( !$Account )
        {
            throw 'Parameter -Accou(n)t has to be specified if -(U)rl parameter is used'
        }
        
        if ( !$Account.Contains( ':' ) )
        {
            throw 'Parameter -Accou(n)t has to specify a value in the format "user:password"'
        }
        
        $accountParts = $Account.split( ':' )
        $credentials = ( New-Object -typename System.Management.Automation.PSCredential -ArgumentList $accountParts[0], ( $accountParts[1] | ConvertTo-SecureString -AsPlainText -Force ) )
        $ws = Use-JobSchedulerWebService -Url $Url -Id $JobSchedulerId -Credentials $credentials
    }

    switch( $What )
    {
        'add' 
        {
            $argument = ''
            
            if ( $EventClass )
            {
                $argument += " -EventClass $EventClass"
            }

            if ( $EventId )
            {
                $argument += " -EventId $EventId"
            }
            
            if ( $ExitCode )
            {
                $argument += " -ExitCode $ExitCode"
            }

            if ( $AllowedExitCodes )
            {
                $argument += " -AllowedExitCodes $AllowedExitCodes"
            }

            if ( $Job )
            {
                $argument += " -Job $Job"
            }

            if ( $JobChain )
            {
                $argument += " -JobChain $JobChain"
            }

            if ( $Order )
            {
                $argument += " -Order $Order"
            }

            if ( $SupervisorJobChain )
            {
                $argument += " -SupervisorJobChain $SupervisorJobChain"
            }

            if ( $ExpirationDate )
            {
                if ( $ExpirationDate -eq 'never' )
                {
                    $argument += " -NoExpiration"
                } else {
                    # $argument += "-ExpirationDate ""$( [DateTime]::ParseExact($ExpirationDate, 'yyyy-MM-dd HH:mm:ss', $null) )"""
                    $argument += " -ExpirationDate $ExpirationDate"
                }
            }

            if ( $ExpirationCycle )
            {
                $argument += " -ExpirationCycle $ExpirationCycle"
            }

            if ( $ExpirationPeriod )
            {
                $argument += " -ExpirationPeriod $ExpirationPeriod"
            }

            $paramsHashtable = @{}
            for( $i=0; $i -lt $Parameters.length; $i++ )
            {
                $parameterParts = $Parameters[$i].split( '=' )
                if ( $parameterParts.length -ne 2 )
                {
                    throw 'illegal value for parameter -Parameters, use "name=value,name=value"'
                }
                
                Write-Debug ".. adding parameter $($parameterParts[0]) = $($parameterParts[1])"
                $paramsHashtable.add( $parameterParts[0], $parameterParts[1] )
            }
            
            if ( $paramsHashtable )
            {
                $argument += " -Parameters " + '$paramsHashtable'
            }
            
            $command = "Add-JobSchedulerEvent $argument"
            Write-Verbose ".. executing cmdlet: $command"
            [Object[]] $events = Invoke-Expression $command
        }

        'remove' 
        {
            $argument = ''

            if ( $EventClass )
            {
                $argument += " -EventClass $EventClass"
            }

            if ( $EventId )
            {
                $argument += " -EventId $EventId"
            }
            
            if ( $JobChain )
            {
                $argument += " -SupervisorJobChain $JobChain"
            }

            if ( $SupervisorJobChain )
            {
                $argument += " -SupervisorJobChain $SupervisorJobChain"
            }

            $command = "Remove-JobSchedulerEvent $argument"
            Write-Verbose ".. executing cmdlet: $command"
            [Object[]] $events = Invoke-Expression $command
        }

        'check' 
        {
            $argument = ''

            if ( $EventClass )
            {
                $argument += " -EventClass $EventClass"
            }

            if ( $EventId )
            {
                $argument += " -EventId $EventId"
            }
            
            if ( $ExitCode )
            {
                $argument += " -ExitCode $ExitCode"
            }

            if ( $JobChain )
            {
                $argument += " -SupervisorJobChain $JobChain"
            }

            if ( $SupervisorJobChain )
            {
                $argument += " -SupervisorJobChain $SupervisorJobChain"
            }

            if ( $XPath )
            {
                $argument += " -XPath $XPath"
            }

            $command = "Get-JobSchedulerEvent $argument"
            Write-Verbose ".. executing cmdlet: $command"
            [Object[]] $events = Invoke-Expression $command
        }
    }

    $events.count